import importlib


class ApplicationModule:
    def __init__(self, name: str):
        self.name = name

    def import_from(self, module_name: str):
        importlib.import_module(module_name)

    def __repr__(self):
        return f"<{self.name} {object.__repr__(self)}>"



from typing import Any

from lato.application_module import ApplicationModule
from lato.dependency_provider import DependencyProvider
from lato.transaction_context import TransactionContext


class Application(ApplicationModule):
    dependency_provider_class = DependencyProvider

    def __init__(self, name=__name__, dependency_provider=None, **kwargs):
        super().__init__(name)
        self.dependency_provider = (
            dependency_provider or self.dependency_provider_class(**kwargs)
        )
        self._on_enter_transaction_context = lambda ctx: None
        self._on_exit_transaction_context = lambda ctx, exception=None: None
        self._transaction_middlewares = []
        self._modules = set([self])

    def include_module(self, a_module):
        assert isinstance(
            a_module, ApplicationModule
        ), "Can only include ApplicationModule instances"
        self._modules.add(a_module)

    def get_dependency(self, identifier: Any) -> Any:
        """Get a dependency from the dependency provider"""
        return self.dependency_provider.get_dependency(identifier)

    def __getitem__(self, item) -> Any:
        return self.get_dependency(item)

    def on_enter_transaction_context(self, func):
        """
        Decorator for registering a function to be called when entering a transaction context

        :param func:
        :return:
        """
        self._on_enter_transaction_context = func
        return func

    def on_exit_transaction_context(self, func):
        """
        Decorator for registering a function to be called when exiting a transaction context

        :param func:
        :return:
        """
        self._on_exit_transaction_context = func
        return func

    def on(self, event_name):
        def decorator(event_handler):
            """
            Decorator for registering an event handler

            :param event_handler:
            :return:
            """
            # TODO: not yet implemented
            return event_handler

        return decorator

    def transaction_middleware(self, middleware_func):
        """
        Decorator for registering a middleware function to be called when executing a function in a transaction context
        :param middleware_func:
        :return:
        """
        self._transaction_middlewares.insert(0, middleware_func)
        return middleware_func

    def transaction_context(self, **dependencies) -> TransactionContext:
        """
        Creates a transaction context with the application dependencies

        :param dependencies:
        :return:
        """
        dp = self.dependency_provider.copy(**dependencies)
        ctx = TransactionContext(dp)
        ctx._on_enter_transaction_context = self._on_enter_transaction_context
        ctx._on_exit_transaction_context = self._on_exit_transaction_context
        ctx._transaction_middlewares = self._transaction_middlewares
        return ctx



from collections import defaultdict
from functools import partial
from typing import Any

from lato.dependency_provider import DependencyProvider


class TransactionContext:
    """A context spanning a single transaction for execution of a function"""

    dependency_provider_factory = DependencyProvider

    def __init__(self, dependency_provider=None, *args, **kwargs):
        self.dependency_provider = (
            dependency_provider or self.dependency_provider_factory(*args, **kwargs)
        )
        self._on_enter_transaction_context = lambda ctx: None
        self._on_exit_transaction_context = lambda ctx, exception: None
        self._transaction_middlewares = []

    def begin(self):
        """Should be used to start a transaction"""
        self._on_enter_transaction_context(self)

    def end(self, exception=None):
        """Should be used to commit/end a transaction"""
        self._on_exit_transaction_context(self, exception)

    def __enter__(self):
        self.begin()
        return self

    def __exit__(self, exc_type=None, exc_val=None, exc_tb=None):
        self.end(exc_val)

    def _wrap_with_middlewares(self, handler_func):
        p = handler_func
        for middleware in self._transaction_middlewares:
            p = partial(middleware, self, p)
        return p

    def _get_overrides(self, **kwargs):
        overrides = dict(ctx=self)
        overrides.update(kwargs)

        type_match = defaultdict(list)
        for name, value in overrides.items():
            type_match[type(value)].append(value)
        with_unique_type = dict((k, v[0]) for k, v in type_match.items() if len(v) == 1)

        overrides.update(with_unique_type)

        return overrides

    def call(self, func, *func_args, **func_kwargs):
        dp = self.dependency_provider.copy(ctx=self)
        resolved_kwargs = dp.resolve_func_params(func, func_args, func_kwargs)
        p = partial(func, **resolved_kwargs)
        wrapped_handler = self._wrap_with_middlewares(p)
        result = wrapped_handler()
        return result

    def emit(self, event: str, *args, **kwargs):
        """Emit an event"""
        # TODO: Implement event emitting

    def get_dependency(self, identifier: Any) -> Any:
        """Get a dependency from the dependency provider"""
        return self.dependency_provider.get_dependency(identifier)

    def __getitem__(self, item) -> Any:
        return self.get_dependency(item)
